# deepsea-ai, Apache-2.0 license
# Filename: __main__
# Description: Main entry point for the deepsea_ai command line interface
import time
from datetime import datetime

import boto3
import click
import os
from pathlib import Path
from urllib.parse import urlparse

from deepsea_ai.config.config import Config
from deepsea_ai.commands import upload_tag, process, train, bucket, monitor
from deepsea_ai.config import config as cfg
from deepsea_ai.config import setup
from deepsea_ai.database.job.database import Job, init_db
from deepsea_ai.database.job.misc import JobType
from deepsea_ai import logger
from deepsea_ai.logger import info, err, debug, warn, critical
from deepsea_ai import __version__
from deepsea_ai import common_args

default_config = cfg.Config(quiet=True)
default_config_ini = cfg.default_config_ini
cfg_option = click.option('--config', type=str, default=default_config_ini,
                          help=f'Path to config file to override defaults in {default_config_ini}')


def init(log_prefix: str = "deepsea_ai", config: str = default_config_ini) -> Config:
    python_path = Path(os.environ.get('LOG_PATH', 'logs'))
    logger.create_logger_file(python_path, log_prefix)

    # get the AWS profile from the environment and use it for all AWS commands
    if 'AWS_DEFAULT_PROFILE' in os.environ or 'AWS_PROFILE' in os.environ:
        if 'AWS_DEFAULT_PROFILE' in os.environ:
            info(
                f'AWS_DEFAULT_PROFILE is set to {os.environ["AWS_DEFAULT_PROFILE"]} and will be used for all AWS commands')
            profile = os.environ['AWS_DEFAULT_PROFILE']
            boto3.setup_default_session(profile_name=profile)
        if 'AWS_PROFILE' in os.environ:
            info(f'AWS_PROFILE is set to {os.environ["AWS_PROFILE"]} and will be used for all AWS commands')
            profile = os.environ['AWS_PROFILE']
            boto3.setup_default_session(profile_name=profile)

    # initialize the config file either from the default or a custom location
    if config:
        custom_config = cfg.Config(config)
    else:
        custom_config = cfg.Config()

    return custom_config


user_name = default_config.get_username()

# example s3 buckets for help
example_input_process_s3 = f's3://{user_name}-video-in-dev'
example_output_process_s3 = f's3://{user_name}-tracks-out-dev'
example_input_train_s3 = f's3://{user_name}-training-dev'
example_output_train_s3 = f's3://{user_name}-model-checkpoints-dev'


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(
    __version__,
    '-V', '--version',
    message=f'%(prog)s, version %(version)s'
)
def cli():
    """
    Process deep sea video in AWS from a command line.
    """
    pass


@cli.command(name="setup")
@click.option('--config', type=str, required=False,
              help=f'Path to config file to override defaults in {default_config_ini}')
@click.option('--mirror', is_flag=True, default=False, help='Mirror docker images from dockerhub into your Elastic '
                                                            'Container Registry (ECR)')
def setup_command(config, mirror):
    """
     Setup your AWS environment. Only need to run this once unless running in a new AWS account.
    """
    init(log_prefix="deepsea_ai_setup")
    custom_config = cfg.Config(config)
    init_db(custom_config)
    account = custom_config.get_account()
    region = custom_config.get_region()
    image_cfg = ['yolov5_container', 'strongsort_container']
    image_tags = [custom_config('docker', t) for t in image_cfg]
    if mirror:
        setup.mirror_docker_hub_images_to_ecr(ecr_client=boto3.client("ecr"), account_id=account,
                                              region=region, image_tags=image_tags)
    setup.create_role(account_id=account)
    setup.store_role(custom_config)

    setup.setup_default_data(custom_config)


@cli.command(name="ecsprocess")
@click.option('-u', '--upload', is_flag=True, default=False,
              help='Set option to upload local video files to S3 bucket')
@click.option('--clean', is_flag=True, default=False,
              help='Clean up unused artifact (e.g. video) from s3 after processing')
@click.option('-i', '--input', type=str, default="/Volumes/M3/mezzanine/DocRicketts/2022/02/1423/",
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg understands. This can also be a single video file.')
@click.option('-e', '--exclude', type=str, multiple=True,
              help='Exclude directory or file. Excludes any directory or file that contains the given string')
@common_args.job_option
@common_args.cluster_option
@common_args.dry_run_option
@cfg_option
@common_args.args
def ecs_process(config, upload, clean, cluster, job, input, exclude, dry_run, args):
    """
     (optional) upload, then batch process in an ECS cluster
    """
    custom_config = init(log_prefix="dsai_ecsprocess", config=config)
    session_maker = init_db(custom_config)
    input_path = Path(input)
    processor = cluster
    video_bucket = 'Unknown'
    resources = None
    if not dry_run:
        resources = custom_config.get_resources(cluster)
        video_bucket = resources['VIDEO_BUCKET']
        processor = resources['PROCESSOR']

    user_name = custom_config.get_username()
    videos = custom_config.check_videos(input_path, exclude)
    tags = custom_config.get_tags(f'Video uploaded from {input} by user {user_name} ')

    total_submitted = 0
    for v in videos:
        loaded = False
        if upload and not loaded:
            if dry_run:
                info(f'Dry run: Uploading {v.name} to S3 bucket {video_bucket}')
            else:
                upload_tag.video_data([v], urlparse(f's3://{video_bucket}'), tags)

        if not loaded:
            if dry_run:
                info(f'Dry run: Submitting {v.name} to cluster for processing with job {job}, cluster {cluster},processor {processor}, user {user_name}, clean {clean}, args {args}')
            else:
                process.batch_run(session_maker, resources, v, job, user_name, clean, args)
            total_submitted += 1
        else:
            warn(f'Video {v.name} has already been processed and loaded...skipping')

    info(f'==== Submitted {total_submitted} videos to {processor} for processing =====')


@cli.command(name="process")
# this might be cleaner as an enum but click does not support that fully yet
@click.option('-i', '--input', type=str, required=True,
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg understands. This can also be a single video file.')
@click.option('-e', '--exclude', type=str, multiple=True,
              help='Exclude directory or file. Excludes any directory or file that contains the given string')
@click.option('--input-s3', type=str, required=True,
              help=f'Path to the s3 bucket with video files. These can be either mp4 or '
                   f'mov files that ffmpeg understands, e.g. s3://{example_input_process_s3}')
@click.option('--output-s3', type=str, required=True,
              help=f'Path to the s3 bucket to store the output, e.g. s3://{example_output_process_s3}')
@click.option('-m', '--model-s3', type=str, default=default_config('aws', 'model'),
              help='S3 location of the trained model tar gz file - must contain a model.tar.gz file with a valid YOLOv5 '
                   'Pytorch model.')
@click.option('--instance-type', type=str, default='ml.g4dn.xlarge',
              help='AWS instance type, e.g. ml.g4dn.xlarge, ml.c5.xlarge')
@common_args.dry_run_option
@common_args.args
@common_args.job_option
@common_args.config_s3_option
@cfg_option
def process_command(config, dry_run, input, exclude, input_s3, output_s3, model_s3, config_s3, job, instance_type, args):
    """
     upload video(s) then process with a model
    """
    custom_config = init(log_prefix="dsai_process", config=config)
    session_maker = init_db(custom_config)

    # get tags to apply to the resources for cost monitoring
    tags = custom_config.get_tags(job)

    input_path = Path(input)
    input_s3 = urlparse(input_s3.rstrip('/'))
    output_s3 = urlparse(output_s3.rstrip('/'))
    model_s3 = urlparse(model_s3.rstrip('/'))

    # create the buckets
    info(f'Creating buckets')

    if bucket.create(input_s3, tags, dry_run) and bucket.create(output_s3, tags, dry_run):

        videos = custom_config.check_videos(input_path, exclude)
        input_s3, size_gb = upload_tag.video_data(videos, input_s3, tags, dry_run)

        # size in GB of the input data should never be < 1
        if size_gb < 1:
            size_gb = 1

        # insert the datetime prefix to make a unique key for the output
        now = datetime.utcnow()
        prefix = now.strftime("%Y%m%dT%H%M%SZ")
        output_unique_s3 = urlparse(f"s3://{output_s3.netloc}/{output_s3.path.lstrip('/')}/{prefix}/")

        # Check if the --save-vid flag is set in the args
        save_vid = False
        if args and 'save_vid' in args:
            save_vid = True

        # estimate the volume size needed for the job; make it 2x the size of the input if saving the video
        if save_vid:
            volume_size_gb = int(2 * size_gb)
        else:
            volume_size_gb = int(1.25 * size_gb)

        # create the arguments for the process script
        process.script_processor_run(session_maker=session_maker,
                                     input_s3=input_s3,
                                     output_s3=output_unique_s3,
                                     model_s3=model_s3,
                                     config_s3=config_s3,
                                     volume_size_gb=volume_size_gb,
                                     instance_type=instance_type,
                                     tags=tags,
                                     dry_run=dry_run,
                                     custom_config=custom_config,
                                     args=args)


@cli.command(name="upload")
@click.option('--config', type=str, default=default_config_ini,
              help=f'Path to config file to override defaults in {default_config_ini}')
@click.option('-i', '--input', type=click.Path(exists=True, file_okay=False, dir_okay=True), required=True,
              help='Path to the folder with video files to upload. These can be either mp4 or mov files that '
                   'ffmpeg understands.  This can also be a single video file.')
@click.option('--s3', type=str, help='S3 bucket to upload to, e.g. s3://902005-video-in-dev', required=True)
def upload_command(config, input, s3):
    """
    Upload videos
    """
    custom_config = init(log_prefix="dsai_upload", config=config)
    init_db(custom_config)
    input_s3 = urlparse(s3.rstrip('/'))
    tags = custom_config.get_tags(f'Uploaded {input} to {s3}')
    bucket.create(input_s3, tags)
    videos = custom_config.check_videos(Path(input))
    upload_tag.video_data(videos, input_s3, tags)


@cli.command(name="train")
@click.option('--config', type=str, default=default_config_ini,
              help=f'Path to config file to override defaults in {default_config_ini}')
@click.option('--images', type=str, required=True,
              help='Path to compressed training images. Images should be put into a tar.gz file '
                   'that decompress into images/train images/val and optionally images/test. This compressed file should not include any archive artifacts,'
                   ' e.g. ._ files. For example, compress to avoid archives in Mac OSX with'
                   ' COPYFILE_DISABLE=1 tar -czf images.tar.gz images/.  ')
@click.option('--labels', type=str, required=True,
              help='Path to the compressed training labels in YOLO format. Labels should be put into a tar.gz file '
                   'that decompress into labels/. This compressed file should not include any archive artifacts,'
                   ' e.g. ._ files. For example, compress to avoid archives in Mac OSX with'
                   ' COPYFILE_DISABLE=1 tar -czf labels.tar.gz labels/.  ')
@click.option('--label-map', type=str, required=True,
              help='Path to a simple text file with the yolo names in the sorted order of the training label indexes')
@click.option('--input-s3', type=str, required=True,
              help=f'Path to the s3 bucket to save the training data, e.g. {example_input_train_s3} or {example_input_train_s3}')
@click.option('--output-s3', type=str, required=True,
              help=f'Path to the s3 bucket to store the output, e.g. {example_output_train_s3} or {example_output_train_s3}')
@click.option('--resume', type=bool, default=False, help="Resume training from previous run")
@click.option('--model', type=str, default='yolov5x', help=f"Model choice: {','.join(train.models)} ")
@click.option('--epochs', type=int, default=2, help='Number of epochs. Default 2.')
@click.option('--batch-size', type=int, default=2, help='Batch size. Default 2.')
@click.option('--instance-type', type=str, default='ml.p3.2xlarge',
              help='AWS instance type, e.g. ml.p3.2xlarge, ml.p3.8xlarge, ml.p3.16xlarge, ml.p4d.24xlarge')
def train_command(config, images, labels, label_map, input_s3, output_s3, resume, model, epochs, batch_size,
                  instance_type):
    """
     (optional) upload training data, then train a YOLOv5 model
    """
    custom_config = init(log_prefix="dsai_train", config=config)
    init_db(custom_config)

    if instance_type == 'ml.p2.xlarge':
        critical(f'{instance_type} too small for model {model}. Choose ml.p3.2xlarge or better')
        raise Exception(f'{instance_type} too small for model {model}. Choose ml.p3.2xlarge or better')

    # get tags to apply to the resources for cost monitoring
    tags = custom_config.get_tags(f'Training {input_s3} with {model} batch {batch_size} instance {instance_type}')

    image_path = Path(images)
    label_path = Path(labels)
    name_path = Path(label_map)

    # strip off any training forward slashes
    input_s3 = urlparse(input_s3.rstrip('/'))
    output_s3 = urlparse(output_s3.rstrip('/'))

    data = [image_path, label_path, name_path]

    # create the buckets
    if bucket.create(input_s3, tags) and bucket.create(output_s3, tags):

        # upload and return the final bucket prefix to the training data and its total size
        input_training, size_gb = upload_tag.training_data(data, input_s3, tags, cfg.default_training_prefix)

        debug(f"Training data: {input_training} size: {size_gb} GB")

        # insert the datetime prefix to make a unique key for the outputs
        now = datetime.utcnow()
        prefix = now.strftime("%Y%m%dT%H%M%SZ")

        if resume:  # resuming from previous bucket, so no need to set prefix
            ckpts_s3 = urlparse(f"s3://{output_s3.netloc}/{output_s3.path.lstrip('/')}")
            volume_size_gb = int(4 * size_gb + 50)
        else:
            ckpts_s3 = urlparse(f"s3://{output_s3.netloc}/{output_s3.path.lstrip('/')}/{prefix}/checkpoints/")
            volume_size_gb = int(2 * size_gb + 50)
        model_s3 = urlparse(f"s3://{output_s3.netloc}/{output_s3.path.lstrip('/')}/{prefix}/models/")

        # train
        train.yolov5(data, input_training, ckpts_s3, model_s3, epochs, batch_size, volume_size_gb, model, instance_type,
                     custom_config)


@cli.command(name="package")
@click.option('--s3', type=str, required=True,
              help=f"Bucket with the model checkpoints, e.g. s3://{example_output_train_s3}/20220821T005204Z"
              )
def package_command(s3):
    """
    Package a YOLOv5 model into a format that the deepsea-ai can use in its pipelines.
    This is done at the end of the train command automatically and stored in a model.tar.gz file.
    This is added in case checkpoints were generated outside the training command, e.g. SageMaker Studio. Colab
    """
    init(log_prefix="dsai_package")
    train.package(urlparse(s3.rstrip('/')))


@cli.command(name="split")
@click.option('-i', '--input', type=str, required=True,
              help='Path to the root folder with images and labels, organized into labels/ and images/ folders files to split')
@click.option('-o', '--output', type=str, required=True,
              help='Path to the root folder to save the split, compressed files. If it does not exist, it will be created.')
def split_command(input: str, output: str):
    """
    Split data into train/val/test sets randomly per the following percentages 85%/10%/5%
    """
    init(log_prefix="dsai_split")
    input_path = Path(input)
    output_path = Path(output)

    if not output_path.exists():
        output_path.mkdir(parents=True)

    paths = [input_path / 'labels', input_path / 'images']

    exists = [not p.exists() for p in paths]
    if any(exists):
        err(f'Error: one or more {paths} missing')
        return

    train.split(Path(input), Path(output))


@cli.command(name="monitor")
@click.option('--cluster', type=str, required=True,
              help='Name of the cluster to query.  This must correspond to an available Elastic '
                   'Container Service cluster.')
@click.option('--config', type=str, required=False,
              help=f'Path to config file to override defaults in {default_config_ini}')
@click.option('--timeout-period', type=int, help='Timeout for monitoring in seconds; default is never')
@click.option('--update-period', type=int, default=10, help='Update period to monitor a job; default is every 60 '
                                                            'seconds. Ignored if --job is not specified.Generates a '
                                                            'new report file in the reports/ folder')
def monitor_command(cluster: str, config, update_period: int, timeout_period: int):
    """
    Print monitoring information for the cluster
    """
    custom_config = init(log_prefix="dsai_monitor", config=config)
    session_maker = init_db(custom_config)
    resources = custom_config.get_resources(cluster)
    if not resources:
        err(f'No resources found for cluster {cluster}')
        return
    report_path = Path.cwd() / 'reports'

    while True:
        with session_maker.begin() as db:
            # Get all the jobs
            num_jobs = len(db.query(Job).filter(Job.job_type == JobType.ECS).all())
            info(f'Found {num_jobs} jobs in the database with type {JobType.ECS}.')

        if num_jobs > 0:
            info(f'Monitoring {num_jobs} job')
            m = monitor.Monitor(session_maker, report_path, resources, update_period)
            m.start()
            m.join(timeout_period)
        else:
            info(f'No jobs found in the database with type {JobType.ECS}. Checking again in 30 seconds. Ctrl-C to stop.')
            time.sleep(30)


if __name__ == '__main__':
    try:
        start = datetime.utcnow()
        cli()
        end = datetime.utcnow()
        info(f'Done. Elapsed time: {end - start} seconds')
    except Exception as e:
        err(f'Exiting. Error: {e}')
        exit(-1)



# deepsea-ai, Apache-2.0 license
# Filename: database/report_generator.py
# Description: Report generator for jobs that are being consumed by an ECS cluster or SageMaker

from pathlib import Path

from deepsea_ai import __version__
from deepsea_ai.config import config as cfg
from deepsea_ai.database.job.database import Job, PydanticJobWithMedias
from deepsea_ai.database.job.misc import job_hash
from deepsea_ai.database.tracks import api, queries
from deepsea_ai.logger import info, debug, err
from datetime import datetime as dt

# Get the model and test data paths
default_config = cfg.Config()


def create_report(job: Job, output_path: Path, resources: dict = None):
    """
    Create a report of the jobs that were run
    :param job: The job
    :param output_path: Path to write the report to
    :param resources: (optional) The resources dictionary of the cluster
    """

    # create the output path if it doesn't exist
    output_path.mkdir(parents=True, exist_ok=True)

    # create a file name that replaces spaces with underscores and adds a timestamp
    job_report_name = f"{job.name.replace(' ', '_')}_{dt.utcnow().strftime('%Y%m%d')}.txt"
    output_path = output_path / job_report_name
    info(f"Creating job report for {job.name} in {output_path}")

    jobs_p = PydanticJobWithMedias.from_orm(job)
    created_time = jobs_p.createdAt
    num_media = len(jobs_p.media)

    job_report_name = f"{jobs_p.name}, Total media: {num_media}, Created at: {created_time} "

    # Get additional information from the deepsea_ai database if it exists
    if default_config('database', 'gql') and resources:
        try:
            database = api.DeepSeaAIClient(default_config('database', 'gql'))
            jobs_q = database.execute(queries.GET_JOB_SUMMARY, job_uuid=job_hash(f"{resources['PROCESSOR']}{job.name}"))
            if len(jobs_q['data']['jobs']) > 0:
                job_id = jobs_q['data']['jobs'][0]['id']
                job_detail = jobs_q['data']['jobs'][0]['detail']
                debug(f"JobCache: Found job id: {job_id}")
                job_report_name += f", Job: {job_id}, {job_detail}"
        except Exception as e:
            err(f"Unable to fetch job id from deepsea_ai database: {e}")

    # Remove the report if it exists
    if output_path.exists():
        output_path.unlink()

    # Write the report
    with open(output_path.as_posix(), 'w') as f:
        f.write(f"DeepSea-AI {__version__}\n")
        f.write(f"Job: {job_report_name}\n")
        f.write(f"==============================================================================================\n")
        f.write(f"Index, Media, Created, Last Updated, Status\n")

        # Write the status of each media file in the job
        names = [m.name for m in jobs_p.media]
        for idx, name in enumerate(sorted(names)):
            # Get the media with the name
            media = [m for m in jobs_p.media if m.name == name][0]
            f.write(f"{idx}, {name}, {media.createdAt}, {media.updatedAt}, {media.status}\n")

    # Write the report in a web-friendly table format, highlighting the status in red if it is not complete
    # with open(output_path.as_posix().replace('.txt', '.html'), 'w') as f:
    #     f.write(f"<html><head><title>DeepSea-AI {__version__}</title></head><body>")
    #     f.write(f"<h1>Job: {job_report_name}</h1>")
    #     f.write(f"<table border=1><tr><th>Index</th><th>Media</th><th>Last Updated</th><th>Status</th></tr>")
    #     for idx, name in enumerate(sorted(names)):
    #         # Get the media with the name
    #         media = [m for m in jobs_p.media if m.name == name][0]
    #         if media.status != 'COMPLETE':
    #             f.write(f"<tr><td>{idx}</td><td>{name}</td><td>{media.updatedAt}</td><td><font color='red'>{media.status}</font></td></tr>")
    #         else:
    #             f.write(f"<tr><td>{idx}</td><td>{name}</td><td>{media.updatedAt}</td><td>{media.status}</td></tr>")
    #     f.write(f"</table></body></html>")



# deepsea-ai, Apache-2.0 license
# Filename: logger/__init__.py
# Description: Logger for deepsea-ai. Logs to both a file and the console.
# Creates a global data frame to store a summary of the results.

import logging
from pathlib import Path
from datetime import datetime as dt

import pandas as pd

LOGGER_NAME = "DSEAAI"
DEBUG = True
keys = ["job", "video", "time", "status", "message"]


class _Singleton(type):
    """ A metaclass that creates a Singleton base class when called. """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(_Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Singleton(_Singleton('SingletonMeta', (object,), {})): pass


class CustomLogger(Singleton):
    logger = None
    output_path = Path.cwd()

    def __init__(self, output_path: Path = Path.cwd(), output_prefix: str = "deepsea_ai"):
        """
        Initialize the logger
        """
        self.logger = logging.getLogger(LOGGER_NAME)
        self.logger.setLevel(logging.DEBUG)
        self.output_path = output_path
        self.output_path.mkdir(parents=True, exist_ok=True)
        formatter = logging.Formatter("%(asctime)s %(levelname)s %(message)s")

        # default log file date to today
        now = dt.utcnow()

        # log to file
        self.log_filename = output_path / f"{output_prefix}_{now:%Y%m%d}.log"
        handler = logging.FileHandler(self.log_filename, mode="w")
        handler.setFormatter(formatter)
        handler.setLevel(logging.INFO)
        self.logger.addHandler(handler)

        # also log to console
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        console.setFormatter(formatter)
        self.logger.addHandler(console)

        self.logger.info(f"Logging to {self.log_filename}")

    def loggers(self) -> logging.Logger:
        return self.logger

def create_logger_file(log_path: Path, prefix: str = "deepsea_ai"):
    """
    Create a logger file
    :param log_path: Path to the log file
    """
    # create the log directory if it doesn't exist
    log_path.mkdir(parents=True, exist_ok=True)
    return CustomLogger(log_path, prefix)


def custom_logger() -> logging.Logger:
    """
    Get the logger
    """
    return logging.getLogger(LOGGER_NAME)


def err(s: str):
    custom_logger().error(s)


def info(s: str):
    custom_logger().info(s)


def debug(s: str):
    custom_logger().debug(s)


def warn(s: str):
    custom_logger().warning(s)


def exception(s: str):
    custom_logger().exception(s)


def critical(s: str):
    custom_logger().critical(s)



# deepsea-ai, Apache-2.0 license
# Filename: config/config.py
# Description: Configuration helper to setup defaults and fetch cloud configuration

import string

import boto3
from configparser import ConfigParser
import datetime as dt
import os

import botocore
from botocore.exceptions import ClientError
from pathlib import Path
from typing import List, Any
from deepsea_ai.logger import err, info, debug, warn, critical, exception

default_training_prefix = 'training'
default_config_ini = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.ini')


class Config:

    def __init__(self, path: str = None, quiet: bool = False):
        """
        Read the .ini file and parse it
        """
        # set the default output path of the sqlite job database to the same directory as the config.ini file
        self.job_db_path = Path(os.path.dirname(os.path.abspath(__file__)))
        self.parser = ConfigParser()
        if path:
            self.path = path
        else:
            self.path = default_config_ini

        if not os.path.isfile(self.path):
            raise Exception(f'Bad path to {self.path}. Is your {self.path} missing?')

        self.parser.read(self.path)
        lines = open(self.path).readlines()
        if not quiet:
            info(f"=============== Config file {self.path} =================")
            for l in lines:
                info(l.strip())

            if not path:
                info(f"============ You can override these settings by creating a customconfig.ini file and pass that "
                     f"in with --config=customconfig.ini =====")

    def __call__(self, *args, **kwargs):
        assert len(args) == 2
        return self.parser.get(args[0], args[1])

    def get_db_path(self) -> Path:
        """
        Get the path to the job database
        :return:
        """
        return self.job_db_path

    def save(self, *args, **kwargs):
        assert len(args) == 3
        self.parser.set(section=args[0], option=args[1], value=args[2])
        with open(self.path, 'w') as fp:
            self.parser.write(fp)

    def get_role(self) -> str:
        """
        Get the user role; first check the environment variable, then the config
        :return role ARN string
        """
        if 'SAGEMAKER_ROLE' in os.environ:
            return os.environ['SAGEMAKER_ROLE']

        sagemaker_arn = self.__call__('aws', 'sagemaker_arn')
        if not sagemaker_arn:
            raise Exception('Run deepsea-ai setup or set the SAGEMAKER_ROLE environment variable')
        return sagemaker_arn

    @staticmethod
    def get_account() -> str:
        """
        Get the account number associated with this user
        :return:
        """
        try:
            account_number = boto3.client('sts').get_caller_identity()['Account']
            info(f'Found account {account_number}')
            return account_number
        except ClientError as e:
            err(e)
            msg = f'Could not get account number from AWS. Check your config.ini file. ' \
                  f'Account number is not set in the config.ini file and AWS credentials are not configured.'
            err(msg)
            return None
        except botocore.exceptions.NoCredentialsError as e:
            err(e)
            return None

    @staticmethod
    def get_region() -> str:
        """
        Get the region associated with this user
        :return:
        """
        session = boto3.session.Session()
        region = session.region_name
        info(f'Found region {region}')
        return region

    @staticmethod
    def get_username() -> str:
        """
        Get the user name using IAM; if IAM is not configured, this will default to the root user which may be the case
        for a new AWS account
        :return:
        """
        try:
            sts = boto3.client('sts')
            response = sts.get_caller_identity()
            user_name = response['Arn'].split("/")[-1].split("@")[0]
        except ClientError as e:
            user_name = "Unknown"
        except botocore.exceptions.NoCredentialsError as e:
            user_name = "Unknown"

        return user_name

    def get_tags(self, description: str) -> dict:
        """
        Configure tag dictionary to associate with any AWS resource.
        This is useful for cost accounting and general reporting
        :param description: descriptive information about the use
        :return dictionary with tag key/value pairs
        """
        organization = self.__call__('tags', 'organization')
        project_number = self.__call__('tags', 'project_number')
        stage = self.__call__('tags', 'stage')
        application = self.__call__('tags', 'application')
        user_name = self.get_username()
        deletion_date = (dt.datetime.utcnow() + dt.timedelta(days=90)).strftime('%Y%m%dT%H%M%SZ')
        tag_dict = [{'Key': f'{organization}:project-number', 'Value': project_number},
                    {'Key': f'{organization}:owner', 'Value': user_name},
                    {'Key': f'{organization}:description', 'Value': description},
                    {'Key': f'{organization}:stage', 'Value': stage},
                    {'Key': f'{organization}:application', 'Value': application},
                    {'Key': f'{organization}:deletion-date', 'Value': deletion_date},
                    {'Key': f'{organization}:created-by', 'Value': user_name}]

        # check that the tags conform to the AWS tagging standard (no spaces, no special characters)
        # iterate over the tag dictionary and check the key and value
        for tag in tag_dict:
            info(f'Checking tag {tag}')
            #  The allowed characters across services are: letters (a-z, A-Z), numbers (0-9),
            #  and spaces representable in UTF-8, and the following characters: + - = . _ : / @.
            allowed_chars = ['+', '-', '=', '.', '_', ':', '/', '@']
            allowed_letters = list(string.ascii_letters)
            allowed_numbers = list(string.digits)
            allowed = allowed_chars + allowed_letters + allowed_numbers
            if not any(char in tag['Value'] for char in allowed):
                msg = f'Tag {tag} has a value with special characters. Check your config.ini file. ' \
                      f'Special characters are not allowed in AWS tags, e.g. dots, etc.'
                err(msg)
                raise (msg)

        return tag_dict

    @staticmethod
    def get_resources(stack_name: str) -> dict:
        """
        Get resources relevant to the pipeline from the stack name
        :param stack_name: name of the stack to query in the ECS cluster
        :return: dictionary with resource names
        """
        client_cf = boto3.client('cloudformation')
        client_ecs = boto3.client('ecs')

        try:
            stack_resources = client_cf.list_stack_resources(StackName=stack_name)
            resources = {'CLUSTER': stack_name}

            # fetch the PROCESSOR, etc. environment variables for the single task in the stack; the job is keyed uniquely to it
            key = ['PROCESSOR', 'TRACK_QUEUE', 'VIDEO_QUEUE', 'DEAD_QUEUE', 'TRACK_BUCKET', 'VIDEO_BUCKET']
            for r in stack_resources['StackResourceSummaries']:
                if 'AWS::ECS::TaskDefinition' in r['ResourceType']:
                    arn = r['PhysicalResourceId']
                    task_def = client_ecs.describe_task_definition(taskDefinition=arn)
                    environment = task_def['taskDefinition']['containerDefinitions'][0]['environment']
                    for e in environment:
                        for k in key:
                            if k in e['name']:
                                resources[k] = e['value']
                    break
                if 'AWS::AutoScaling::AutoScalingGroup' in r['ResourceType']:
                    resources['ASG'] = r['PhysicalResourceId']
            return resources
        except ClientError as ex:
            exception(ex)
            if 'AccessDenied' in ex.response['Error']['Code']:
                critical('Access denied; verify you are using the correct AWS credentials')
                raise Exception('Access denied')
            if 'ExpiredToken' in ex.response['Error']['Code']:
                critical('Token expired; you need to re-authenticate your AWS credentials')
                raise Exception('Token expired')
            else:
                critical(f'Unknown error: {ex}')
                raise Exception(f'Unknown error: {ex}')
        except botocore.exceptions.NoCredentialsError as ex:
            exception(ex)
            critical('No credentials; verify you have AWS credentials configured')
            return None

        return None

    @staticmethod
    def check_videos(input_path: Path, exclude: tuple) -> List[Path]:
        """
         Check for videos with acceptable suffixes and return the Paths to them
        :param input_path: input path to search (non-recursively) or a single video file
        :param exclude: directory or files to exclude from the list of videos to process
        :return:
        """
        vid_formats = ['.mov', '.avi', '.mp4', '.mpg', '.mpeg', '.m4v', '.wmv', '.mkv']  # acceptable video suffixes

        # convert exclude tuple to list
        excludes = list(exclude)
        if len(excludes) > 0:
            info(f'Excluding any video file or directory that contains {excludes}')
        else:
            info(f'No video file exclusions specified')

        def search(x: Path):
            if excludes:
                found = [x.name.__contains__(e) for e in excludes]
                return (True not in found and x.suffix.lower() in vid_formats and '._' not in x.name)
            else:
                return (x.suffix.lower() in vid_formats and '._' not in x.name)

        # if the input path is a directory, search for videos
        videos = []
        if input_path.is_dir():
            videos = [x for x in input_path.glob("**/*") if search(x)]
        else:
            if search(input_path):
                videos = [input_path]
        num_videos = len(videos)
        info(f'Found {num_videos} videos to process')
        if num_videos == 0:
            err(f'No videos found in {input_path}')

        # Check that the videos are not empty and that they exist
        for video in videos:
            if video.stat().st_size == 0:
                err(f'Video {video} is empty')
                videos.remove(video)

            if not video.exists():
                err(f'Video {video} does not exist')
                videos.remove(video)

        assert (num_videos > 0), "No videos to process"
        return videos



# deepsea-ai, Apache-2.0 license
# Filename: pipeline/run_strongsort.py
# Description: Runs strongsort tracking algorithm on YOLOv5 models
import click
import datetime
import boto3
import tempfile
import json
import signal
import tarfile
import os
import shutil
import subprocess
import sys
from pathlib import Path
from urllib.parse import urlparse
import uuid
from pipeline import queue_processor
from pipeline import __version__

# If running in AWS, we must define the inputs/outputs per the spec
from pipeline.data_models import generate_uuids, parse_events

if 'AWS_CONTAINER_CREDENTIALS_RELATIVE_URI' in os.environ:
    default_input = '/opt/ml/processing/input'
    default_output = '/opt/ml/processing/output'
else:
    default_input = Path(__file__).parent.parent / 'test' / 'in'
    default_output = Path(__file__).parent.parent / 'test' / 'out'

processing_job_cfg_path = Path('/opt/ml/config/processingjobconfig.json')
stop_flag = False

# A known pretrained model
default_model_s3 = 's3://902005-public/models/yolov5x_mbay_benthic_model.tar.gz'  # yolov5 model

MAX_TIMEOUT_SECS = 3 * 60 * 60  # maximum time a file should take to process is 3 hours


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(
    __version__,
    '-V', '--version',
    message=f'%(prog)s, version %(version)s'
)
def cli():
    """
    StrongSort Detect and Track
    """


@cli.command(name="dettrack")
@click.option('-c', '--config-s3', type=click.STRING, help='Location of strongsort tracking algorithm config yaml file')
@click.option('-r', '--reid-weights', type=click.STRING, help='Location to the reid weights')
@click.option('-i', '--input', type=click.STRING, default=default_input,
                                help='Path to video files. These can be either mp4 or mov files that ffmpeg understands.')
@click.option('-o', '--output', type=click.STRING, default=default_output,
                                help='Path to the output to save the results')
@click.option('-m', '--model-s3', type=str, default=default_model_s3,
              help='S3 path to the trained model tar gz file - must contain a valid YOLOv5 Pytorch model. ')
@click.option('--debug', is_flag=True, help='Debugging flag. Skips processing and downloading of the model.')
@click.option('--args', type=str, help='Additional arguments to pass to strong sort track.py script')
def process_command(config_s3, reid_weights, input, output, model_s3, debug, args):
    """
    Process a collection of videos either from an input folder, or from a sqs queue
    """
    signal.signal(signal.SIGTERM, sigterm_handler)
    global stop_flag
    print(f'{__version__}')
    print(f'Processing videos in {input}. ')
    input_path = Path(input)
    output_path = Path(output)
    output_path.parent.mkdir(parents=True, exist_ok=True)
    processor = queue_processor.QueueProcessor(input_path, output_path)

    # download and setup the model once in a temp dir
    download_fini = False
    with tempfile.TemporaryDirectory() as model_temp_dir:

        # strongsort track.py requires separate input/output so let's create that here
        with tempfile.TemporaryDirectory() as in_tmp_dir:
            in_tmp_path = Path(in_tmp_dir) / 'in'
            in_tmp_path.mkdir()
            out_tmp_path = Path(in_tmp_dir) / 'out'
            out_tmp_path.mkdir()

            while processor.has_video() and not stop_flag:
                try:
                    # start fetching new video
                    video = processor.next(out_tmp_path, debug)

                    if video is None:
                        break

                    # only download the model if we have a video to process
                    if not debug and not download_fini:
                        download_fini = True
                        model_path, config_path = download_config(model_s3, config_s3, Path(model_temp_dir))

                    # output tar to save the results to
                    out_tar_path = Path(output_path) / f'{video.stem}.tracks.tar.gz'
                    out_tar_path.parent.mkdir(parents=True, exist_ok=True)

                    start_utc = datetime.datetime.utcnow()

                    # Strip off the quotes
                    if args:
                        args = args.strip('"')

                    if not debug:
                        cmd_track = f'python3 track.py ' \
                                    f'--source {video.input_path} ' \
                                    f'--project {in_tmp_path} ' \
                                    f'--name {video.input_path.stem} ' \
                                    f'--save-txt ' \
                                    f'{args} '
                        if model_path:
                            cmd_track += f'--yolo-weights {model_path} '
                        if reid_weights:
                            cmd_track += f'--strong-sort-weights {reid_weights} '
                        if config_path:
                            cmd_track += f'--config-strongsort {config_path} '
                    else:
                        cmd_track = [f"echo track {input_path} && sleep 60"]

                    print(f'Running {cmd_track}')
                    processor.message_run(video)
                    track_proc = subprocess.Popen(cmd_track, shell=True)
                    track_proc.wait(timeout=MAX_TIMEOUT_SECS)

                    if track_proc.returncode != 0:
                        processor.message_fail(video, track_proc.stdout)
                        continue

                    track_path = in_tmp_path / video.input_path.stem
                    track_path.mkdir(parents=True, exist_ok=True)

                    # copy any configuration files to further downstream data loading/processing
                    if processing_job_cfg_path.exists():
                        print(f'Copying {processing_job_cfg_path} to {track_path}')
                        shutil.copy2(processing_job_cfg_path.as_posix(), track_path.as_posix())
                    else:
                        # create a new job file in the temp_dir with job metadata
                        with open(f"{track_path.as_posix()}/processingjobconfig.json", "w", encoding="utf-8") as j:
                            image_uri = os.environ[
                                'IMAGE_URI'] if 'IMAGE_URI' in os.environ else "mbari/strongsort-yolov5:latest"
                            processor_name = os.environ[
                                'PROCESSOR'] if 'PROCESSOR' in os.environ else "strongsort-yolov5"
                            json.dump({
                                "UserName": video.user_name,
                                "ProcessingJobName": f"{processor_name}-{video.job_name}",
                                "AppSpecification": {
                                    "ImageUri": image_uri,
                                    "ContainerArguments": sys.argv
                                }
                            }, j)

                    # insert the video path into the processing job config file
                    with open(f"{track_path.as_posix()}/processingjobconfig.json", "r", encoding="utf-8") as f:
                        json_dict = json.load(f)

                    json_dict['VideoName'] = video.input_path.name

                    with open(f"{track_path.as_posix()}/processingjobconfig.json", "w", encoding="utf-8") as j:
                        json.dump(json_dict, j, indent=4)

                    # capture the number of unique tracks
                    unique_track_ids = set()

                    # convert the yolo output to a more friendly json output
                    # yolo output is a simple .txt file with the same file prefix as the video_path,
                    # e.g. myvideo.mov output is myvideo.txt
                    yolo_results = track_path / 'tracks' / f'{video.input_path.stem}.txt'

                    # handle missing data
                    if yolo_results.exists():

                        events_sans_uuids = parse_events(yolo_results.as_posix())
                        events = generate_uuids(events_sans_uuids)

                        # save events aggregated by frame in the same format as deepsea-track to simplify loading
                        visual_events = []
                        last_frame = -1
                        for e in events:
                            if last_frame == -1:
                                last_frame = e['frameNum']

                            if e['frameNum'] > last_frame:
                                with open(f"{track_path}/f{last_frame:06}.json", 'w', encoding='utf-8') as f:
                                    json.dump(["visualevents", visual_events], f)
                                    visual_events = []

                            last_frame = e['frameNum']

                            visual_events.append(["visualevent",
                                                  {
                                                      'bounding_box': {
                                                          'height': e['height'],
                                                          'width': e['width'],
                                                          'x': e['x'],
                                                          'y': e['y']
                                                      },
                                                      'class_name': e['name'],
                                                      'confidence': e['confidence'],
                                                      'frame_num': e['frameNum'],
                                                      'occlusion': e['occlusion'],
                                                      'surprise': e['surprise'],
                                                      'track_uuid': e['track_uuid'],
                                                      'uuid': str(uuid.uuid4()),
                                                  }
                                                  ])
                            unique_track_ids.add(e['track_uuid'])

                            if len(visual_events) > 0:
                                with open(f"{track_path}/f{last_frame:06}.json", 'w',
                                          encoding='utf-8') as f:
                                    json.dump(["visualevents", visual_events], f)

                    # make the output unique with a timestamp
                    total_time = datetime.datetime.utcnow() - start_utc
                    processor.message_save(track_path, out_tar_path, video, len(unique_track_ids), total_time.total_seconds())

                    print(f'Captured {len(unique_track_ids)} in {video.name}. Total processing time {total_time}. Started at {start_utc}')

                    # clean temp input dir and remove output
                    for i in in_tmp_path.glob('*'):
                        if not i.is_dir(): i.unlink()
                    for o in out_tmp_path.glob('*'):
                        o.unlink()
                    processor.clean(video)

                except Exception as ex:
                    print(f'System failure exception {ex}')
                    exit(-1)
    print('Done')


def download_s3_file(bucket: str, key: str, local_path: Path) -> bool:
    """
    Download a file from s3 to a local path
    :param bucket:  Bucket name
    :param key: prefix/key to the object
    :param local_path: Local path to save the file
    :return:  True if successful, False otherwise
    """
    try:
        print(f'Downloading s3://{bucket}/{key} to {local_path.as_posix()}')
        if 'AWS_DEFAULT_PROFILE' in os.environ:
            print(f'Using AWS profile {os.environ["AWS_DEFAULT_PROFILE"]}')
            session = boto3.Session(profile_name=os.environ['AWS_DEFAULT_PROFILE'])
            s3 = session.resource('s3')
        else:
            s3 = boto3.resource('s3')
        s3.Bucket(bucket).download_file(key, local_path.as_posix())
        return True
    except Exception as ex:
        print(f'Exception {ex}')
        return False


def download_config(model_uri: str, track_uri: str, output_dir: Path):
    """
    Downloads and unpacks the model data and optionally a tracker config yaml

    :param model_uri: full path to model bucket object
    :param track_uri: full path to track bucket object or None if using default
    :param output_dir: full path to directory to save the output
    :return:  a Path object with the path to the model, path to the tracker config
    e.g. /tmp/9sdfggg/best.pt,  /tmp/9sdfggg/strongsort_config.yaml
    """
    # download the model
    parsed_track_config = None
    if track_uri: parsed_track_config = urlparse(track_uri, allow_fragments=False)
    parsed_url_model = urlparse(model_uri, allow_fragments=False)

    found_config = False
    model_out_path = output_dir / Path(parsed_url_model.path).name
    track_out_yaml = output_dir / 'strongsort_config.yaml'

    try:
        if not download_s3_file(parsed_url_model.netloc, parsed_url_model.path.lstrip('/'), model_out_path):
            raise Exception(f'Cannot find model {model_uri}')

        # download the tracker config
        if parsed_track_config:
            if not download_s3_file(parsed_track_config.netloc, parsed_track_config.path.lstrip('/'), track_out_yaml):
                raise Exception(f'Cannot find tracker config {parsed_track_config}')
            found_config = True

        # If model is a tar file, unpack it
        if model_out_path.suffix == '.gz' or model_out_path.suffix == '.tar':
            print(f'Unpacking model {model_out_path}...')
            tarfile.open(model_out_path).extractall(output_dir)
            print(os.listdir(output_dir))
            # Find the name of the .pt file and use that as the model
            for f in Path(output_dir).rglob('*.pt'):
                if f.suffix == '.pt':
                    model_out_path = output_dir / f
                    break

        # Error if we don't have a model with a .pt extension
        if model_out_path.suffix != '.pt':
            raise Exception(f'Cannot find model {model_out_path}')

        return model_out_path, track_out_yaml if found_config else None

    except Exception as ex:
        print(ex)
        exit(-1)


def sigterm_handler(signal, frame):
    print(f'Got SIGTERM {signal}')
    global stop_flag
    stop_flag = True


if __name__ == '__main__':
    cli()



# deepsea-ai, Apache-2.0 license
# Filename: config/setup.py
# Description: Setup utility for mirror docker images to ECR, create role, and setup default data
###########################################################################################################
# Credit to Alex for his blog: https://alexwlchan.net/2020/11/copying-images-from-docker-hub-to-amazon-ecr/
# and code detailing how to mirror repositories below
###########################################################################################################
import base64
import json
import subprocess
import tempfile
from pathlib import Path
from urllib.parse import urlparse

import boto3
from botocore.exceptions import ClientError
from deepsea_ai.config import config as cfg
from deepsea_ai.config.config import Config
from deepsea_ai.logger import err, info, warn, exception


def get_ecr_repo_names_in_account(ecr_client, *, account_id):
    """
    Returns a set of all the ECR repository names in an AWS account.
    """
    repo_names = set()

    paginator = ecr_client.get_paginator("describe_repositories")
    for page in paginator.paginate(registryId=account_id):
        for repo in page["repositories"]:
            repo_names.add(repo["repositoryName"])

    return repo_names


def docker_login_to_ecr(ecr_client, *, account_id):
    """
    Authenticate Docker against the ECR repository in a particular account.

    The authorization token obtained from ECR is good for twelve hours, so this
    function is cached to save repeatedly getting a token and running `docker login`
    in quick succession.
    """
    response = ecr_client.get_authorization_token(registryIds=[account_id])

    try:
        auth = response["authorizationData"][0]
    except (IndexError, KeyError):
        raise RuntimeError("Unable to get authorization token from ECR!")

    auth_token = base64.b64decode(auth["authorizationToken"]).decode()
    username, password = auth_token.split(":")

    cmd = [
        "docker",
        "login",
        "--username",
        username,
        "--password",
        password,
        auth["proxyEndpoint"],
    ]

    subprocess.check_call(cmd)


def create_ecr_repository(ecr_client, *, name):
    """
    Create a new ECR repository.
    """
    try:
        ecr_client.create_repository(repositoryName=name)
    except ClientError as err:
        if err.response["Error"]["Code"] == "RepositoryAlreadyExistsException":
            exception(err)
            pass
        else:
            raise


def docker(*args):
    """
    Shell out to the Docker CLI.
    """
    subprocess.check_call(["docker"] + list(args))


def mirror_docker_hub_images_to_ecr(ecr_client, *, account_id, region, image_tags):
    """
    Given the name/tag of images in Docker Hub, mirror those images to ECR.
    """

    info(f"Creating all ECR repositories in account {account_id}...")
    existing_repos = get_ecr_repo_names_in_account(ecr_client, account_id=account_id)

    mirrored_repos = set(tag.split(":")[0] for tag in image_tags)
    missing_repos = mirrored_repos - existing_repos

    for repo_name in missing_repos:
        ecr_client.create_repository(repositoryName=repo_name)

    info(f"Authenticating Docker with ECR for account {account_id}...")
    docker_login_to_ecr(ecr_client, account_id=account_id)

    for hub_tag in image_tags:
        ecr_tag = f"{account_id}.dkr.ecr.{region}.amazonaws.com/{hub_tag}"
        info(f"Mirroring {hub_tag} to {ecr_tag}")
        docker("pull", hub_tag)
        docker("tag", hub_tag, ecr_tag)
        docker("push", ecr_tag)


def create_role(account_id: str):
    """
    Sets up the IAM role called DeepSeaAI to support ECR and SageMaker
    :param account_id: AWS account ID
    """
    session = boto3.session.Session()
    iam = session.client('iam')

    path = '/'
    role_name = 'DeepSeaAI'
    description = 'DeepSeaAI Role'
    session_secs = 43200

    ASSUME_ROLE_POLICY_JSON = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": "ecs-tasks.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            },
            {
                "Sid": "",
                "Effect": "Allow",
                "Principal": {
                    "Service": "sagemaker.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            },
            {
                "Sid": "TrustPolicyStatementThatAllowsEC2ServiceToAssumeTheAttachedRole",
                "Effect": "Allow",
                "Principal": {
                    "Service": "ec2.amazonaws.com"
                },
                "Action": "sts:AssumeRole"
            }
        ]
    }

    ROLE_PERMISSIONS_JSON = {
        "Version": "2012-10-17",
        "Statement": [{
            "Effect": "Allow",
            "Action": [
                "iam:GetRole",
                "iam:PassRole"
            ],
            "Resource": f"arn:aws:iam::{account_id}:role/{role_name}"
        }]
    }

    try:

        policy = f"{role_name}GetAndPassRolePolicy"

        iam = session.client('iam')

        iam.create_policy(
            PolicyName=policy,
            PolicyDocument=json.dumps(ROLE_PERMISSIONS_JSON)
        )

        iam.create_role(
            Path=path,
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(ASSUME_ROLE_POLICY_JSON),
            Description=description,
            MaxSessionDuration=session_secs
        )

        iam.attach_role_policy(
            PolicyArn='arn:aws:iam::aws:policy/AmazonSageMakerFullAccess',
            RoleName=role_name
        )
        iam.attach_role_policy(
            PolicyArn='arn:aws:iam::aws:policy/AmazonS3FullAccess',
            RoleName=role_name
        )
        iam.attach_role_policy(
            PolicyArn=f"arn:aws:iam::aws:policy/{policy}",
            RoleName=role_name
        )

    except ClientError as err:
        if err.response["Error"]["Code"] == "EntityAlreadyExists":
            warn(f"Role {role_name} already exists.")
            pass
        else:
            raise


def store_role(config: cfg):
    """
    Save the role in the config
    """
    session = boto3.session.Session()
    iam = session.client('iam')
    results = iam.get_role(RoleName="DeepSeaAI")
    role_arn = results['Role']['Arn']
    info(f'Setting SageMaker Role ARN to {role_arn} in {config.path}')
    config.save('aws', 'sagemaker_arn', role_arn)
    # Parse the ARN to get the account ID
    account_id = role_arn.split(':')[4]
    config.save('aws', 'account_id', account_id)


def download_s3_file(bucket: str, key: str, local_path: Path) -> bool:
    """
    Download a file from s3 to a local path
    :param bucket:  Bucket name
    :param key: prefix/key to the object
    :param local_path: Local path to save the file
    :return:  True if successful, False otherwise
    """
    try:
        info(f'Downloading s3://{bucket}/{key} to {local_path.as_posix()}')
        s3 = boto3.resource('s3', region_name='us-west-2')
        s3.Bucket(bucket).download_file(key, local_path.as_posix())
        return True
    except Exception as ex:
        print(f'Exception {ex}')
        return False


def make_bucket(bucket: str, region: str) -> bool:
    """
    Create a bucket if it doesn't exist
    :param bucket: Bucket name
    :param region: Region name
    :return: True if successful, False otherwise
    """
    s3 = boto3.client('s3', region_name=region)
    try:
        info(f'Creating bucket {bucket} in {region}')
        s3.create_bucket(Bucket=bucket, CreateBucketConfiguration={'LocationConstraint': region})
        return True
    except ClientError as err:
        if err.response["Error"]["Code"] == "BucketAlreadyOwnedByYou":
            warn(f"Bucket {bucket} already exists.")
            return True
        else:
            raise


def setup_default_data(config: Config):
    """
    Setup default data, including video, models, and  track config.
    This creates default data in the following buckets:
    deepsea-ai-<account>-models
    deepsea-ai-<account>-track-conf
    deepsea-ai-<account>-videos
    deepsea-ai-<account>-tracks
    """
    default_model_s3 = config('aws_public', 'model')
    default_track_s3 = config('aws_public', 'track_config')
    account = config.get_account()

    # Create the buckets if they don't exist
    buckets = [f'deepsea-ai-{account}-models',
               f'deepsea-ai-{account}-track-conf',
               f'deepsea-ai-{account}-videos',
               f'deepsea-ai-{account}-tracks']
    for bucket in buckets:
        if not make_bucket(bucket, 'us-west-2'):
            err(f'Cannot create bucket {bucket}')
            return

    # Save the buckets in the config
    config.save('aws', 'models', f's3://{buckets[0]}')
    config.save('aws', 'track_config', f's3://{buckets[1]}')
    config.save('aws', 'videos', f's3://{buckets[2]}')
    config.save('aws', 'tracks', f's3://{buckets[3]}')

    # Download data from S3 to a temp directory and then upload to the new buckets
    with tempfile.TemporaryDirectory() as temp_dir:

        ###########################################################################################
        # Handle video defaults
        ###########################################################################################
        for v in ['video_ex1', 'video_ex2', 'video_ex3']:
            # Parse the s3 path of the default to download
            src = urlparse(config('aws_public', v))
            tgt = urlparse(f's3://{buckets[2]}/{Path(src.path).name}')

            # Skip if the file already exists
            s3 = boto3.resource('s3')
            bucket_tgt = s3.Bucket(tgt.netloc)
            found = False
            for obj in bucket_tgt.objects.all():
                if Path(src.path).name in obj.key:
                    info(f'Found {Path(src.path).name} in {tgt.netloc} so skipping')
                    found = True

            if not found:
                # Download the file from the source bucket
                out_path = Path(temp_dir) / Path(src.path).name
                if not download_s3_file(src.netloc, src.path.lstrip('/'), out_path):
                    err(f'Cannot find {s3}')
                    return
                else:
                    info(f'Uploading {out_path.as_posix()} to {tgt.netloc}')
                    s3.Bucket(tgt.netloc).upload_file(out_path.as_posix(), out_path.name)

        ###########################################################################################
        # Handle model and track config defaults
        ###########################################################################################

        # Dictionary of defaults and their associated buckets to upload to
        defaults = {default_model_s3: f"s3://{buckets[0]}", default_track_s3: f"s3://{buckets[1]}"}

        for bucket_source, bucket_target in defaults.items():
            # Parse the s3 path of the default to download
            src = urlparse(bucket_source)
            tgt = urlparse(bucket_target)

            # Skip if the file already exists
            s3 = boto3.resource('s3')
            bucket_tgt = s3.Bucket(tgt.netloc)
            found = False
            for obj in bucket_tgt.objects.all():
                if Path(src.path).name in obj.key:
                    info(f'Found {Path(src.path).name} in {bucket_target} so skipping')
                    if 'model' in bucket_target:
                        config.save('aws', 'model', f'{bucket_target}/{Path(src.path).name}')
                    elif 'track' in bucket_target:
                        config.save('aws', 'track_config', f'{bucket_target}/{Path(src.path).name}')
                    found = True

            if not found:
                # Download the file from the source bucket
                out_path = Path(temp_dir) / Path(src.path).name
                if not download_s3_file(src.netloc, src.path.lstrip('/'), out_path):
                    err(f'Cannot find {s3}')
                    return
                else:
                    info(f'Uploading {out_path.as_posix()} to {bucket_target}')
                    s3.Bucket(tgt.netloc).upload_file(out_path.as_posix(), out_path.name)


if __name__ == "__main__":
    default_config = cfg.Config()
    default_config_ini = cfg.default_config_ini
    account = default_config.get_account()
    region = default_config.get_region()
    image_cfg = ['yolov5_container', 'strongsort_container']
    image_tags = [default_config('docker', t) for t in image_cfg]
    mirror_docker_hub_images_to_ecr(ecr_client=boto3.client("ecr"), account_id=account, region=region,
                                    image_tags=image_tags)
    create_role(account_id=account)
    store_role(default_config)
    setup_default_data(default_config)



# deepsea-ai, Apache-2.0 license
# Filename: commands/bucket.py
# Description: S3 bucket utilities

import logging
import boto3
import numpy as np
from botocore.exceptions import ClientError
from deepsea_ai.logger import info


def create(bucket: tuple, tags: dict, dry_run: bool = False):
    """Create an S3 bucket
    :param bucket: Bucket to create
    :param tags: Tags to assign to the bucket
    :return: True if bucket created, or it already exists, else False
    """
    if dry_run:
        info(f'dry-run: Would have created bucket {bucket.netloc} with tags {tags}')
        return True

    # Create bucket
    try:
        info(f'Creating bucket {bucket.netloc}...')
        session = boto3.session.Session()
        s3_client = boto3.client('s3', region_name=session.region_name)
        location = {'LocationConstraint': session.region_name}
        s3_client.create_bucket(Bucket=bucket.netloc,
                                CreateBucketConfiguration=location)

        try:
            s3_client.put_bucket_tagging(Bucket=bucket.netloc, Tagging={'TagSet': tags})
        except Exception as error:
            raise f'Error creating bucket {bucket.netloc} {error}'
    except ClientError as e:
        code = e.response['Error']['Code']
        if code == "BucketAlreadyOwnedByYou" or code == 'BucketAlreadyExists':
            logging.info(e)
            return True

        logging.error(e)
        return False
    return True


def size(bucket: tuple) -> int:
    """
    Get the total size in GB of the bucket.
    :param bucket: Bucket name to searc
    :return: Total size in gigabytes
    """
    size_gb = 0
    session = boto3.session.Session()
    s3 = session.resource('s3')
    b = s3.Bucket(bucket.netloc)

    for object in b.objects.filter(Prefix=bucket.path.split('/')[-1]):
        object.key.split('/')[0]
        size_gb += object.size

    return max(np.round(size_gb / 1e9), 1)



# deepsea-ai, Apache-2.0 license
# Filename: commands/monitor.py
# Description: Thread to monitor the status of processing jobs in an ECS cluster.

import time
from pathlib import Path
from threading import Thread

from sqlalchemy.orm import sessionmaker

from deepsea_ai.commands.monitor_utils import log_scaling_activities, log_queue_status
from deepsea_ai.database.job.misc import JobType
from deepsea_ai.database.report_generator import create_report
from deepsea_ai.logger import info, warn, err
from deepsea_ai.database.job.database import Job, Media
from deepsea_ai.database.job.database_helper import json_b64_decode

default_update_period = 60 * 30  # 30 minutes


class Monitor(Thread):
    """
    A new single threaded executor to run the monitor and update the status.
    Run one per cluster.
    """

    def __init__(self, session_maker: sessionmaker,
                 report_path: Path,
                 resources: dict,
                 update_period: int = default_update_period,
                 sim: bool = False):
        """
        :param session_maker: Session maker to connect to the database
        :param report_path: Path to save the report
        :param resources: Dictionary of resources in the cluster
        :param update_period: Period to update the status of jobs messages in the queues
        :param sim: If true, simulate the monitor
        """
        Thread.__init__(self)
        self.report_path = report_path
        self.resources = resources
        self.update_period = update_period
        self.sim = sim
        self.session_maker = session_maker

        info(f'Creating report path {self.report_path} if it does not exist.')
        self.report_path.mkdir(parents=True, exist_ok=True)

    def run(self):
        try:
            if self.sim:
                with self.session_maker() as db:
                    # check the status of the job in the database
                    job = db.query(Job).first()
                    if job:
                        create_report(job, Path('reports'), self.resources)
                return

            # Run forever until Ctrl-C
            while True:
                num_activities = log_scaling_activities(self.resources, num_records=10)

                queue_dict = log_queue_status(self.session_maker, self.resources)
                queue_activity = sum([int(i) for i in queue_dict.values()])

                if num_activities == 0 and queue_activity == 0:
                    info(f'No activity for {self.resources["PROCESSOR"]}.')
                else:
                    # Generate a report every update_period when there is activity
                    if queue_activity > 0:
                        info(f"Getting all media being processed in cluster {self.resources['CLUSTER']}")

                        with self.session_maker() as db:
                            # Get all jobs with the job type ECS
                            jobs_in_clusters = db.query(Job).filter(Job.job_type == JobType.ECS).all()

                            # Get all media in the jobs that are in the cluster
                            for job in jobs_in_clusters:
                                medias = db.query(Media).filter(Media.job_id == job.id).all()
                                info(f"Found {len(medias)} media in job {job.name}")
                                if len(medias) > 0:  # if there are media in the cluster, create a report
                                    create_report(job, Path('reports'), self.resources)

                info(f'Checking again in {self.update_period} seconds. Ctrl-C to stop.')
                time.sleep(self.update_period)
        except KeyboardInterrupt:
            pass
        except Exception as e:
            err(e)
            raise e


# deepsea-ai, Apache-2.0 license
# Filename: database/job/database_helper.py
# Description: Job database

import json
import base64
from datetime import datetime
from sqlalchemy.orm import Session

from deepsea_ai.database.job.database import Job, PydanticJobWithMedias, Media
from deepsea_ai.database.job.misc import Status
from deepsea_ai.logger import info


def json_b64_encode(obj):
    """
    Convert a JSON object to a base64 encoded string
    :param obj: The JSON object to convert
    :return: The base64 encoded string
    """
    json_str = json.dumps(obj)
    encoded = base64.b64encode(json_str.encode()).decode()
    return encoded


def json_b64_decode(obj):
    """
    Decode a base64 encoded JSON string
    :param obj: The base64 encoded JSON string
    :return: The decoded JSON object
    """
    decoded = base64.b64decode(obj).decode()
    return json.loads(decoded)


def get_status(job: Job) -> bool:
    """
    Get the status of a job
    :param job: The job to get the status of
    :return: The status of the job
    """

    # Get the status of all the medias
    statuses = [m.status for m in job.media]

    # if any of the medias are RUNNING, the job should be RUNNING
    if Status.RUNNING in statuses:
        return Status.RUNNING

    # if any the medias are QUEUED, the job should be QUEUED
    if Status.QUEUED in statuses:
        return Status.QUEUED

    # if any of the statuses are FAILED, the job should be FAILED
    if Status.FAILED in statuses:
        return Status.FAILED

    # if all are SUCCESS, the job should be SUCCESS
    num_success = statuses.count(Status.SUCCESS)
    if num_success == len(statuses):
        return Status.SUCCESS

    return Status.UNKNOWN


def get_num_failed(job: Job) -> int:
    """
    Get the number of failed medias in a job
    :param job: The job to get the number of failed medias from
    :return: The number of failed medias in the job
    """

    # Get the status of all the medias
    statuses = [m.status for m in job.media]

    # Count the number of FAILED medias
    num_failed = statuses.count(Status.FAILED)

    return num_failed


def get_num_completed(job: Job) -> int:
    """
    Get the number of completed medias in a job
    :param job: The job to get the number of failed medias from
    :return: The number of completed medias in the job
    """

    # Get the status of all the medias
    statuses = [m.status for m in job.media]

    # Count the number of SUCCESS medias
    num_completed = statuses.count(Status.SUCCESS)

    return num_completed


def get_job_by_name(db: Session, job_name: str) -> Job:
    """
    Get a job from the database by its name
    """
    # Get all the jobs with the same name
    return db.query(Job).filter(Job.name == job_name).first()


def get_job_by_uuid(db: Session, job_uuid: str) -> Job:
    """
    Get a job from the database by its uuid
    """
    # Get all the jobs with the same uuid
    return db.query(Job).filter(Job.uuid == job_uuid).first()


def update_media(db: Session, job: Job, video_name: str, status: str, **kwargs):
    """
    Update a video in a job. If the video does not exist, add it to the job.
    :param db: The database session
    :param job: The job
    :param video_name: The name of the video to update
    :param status: The status of the video
    """
    info(f'Updating media {video_name} to {status}')

    # Set kwargs to empty dict if None
    kwargs = kwargs or {}

    # If there are additional kwargs, search by them and the name
    media = None
    if kwargs:
        if 'metadata_b64' in kwargs:
            # Find the media with the matching metadata
            media = [m for m in job.media if m.metadata_b64 == kwargs['metadata_b64'] and m.name == video_name]
        else:
            for key, value in kwargs.items():
                for m in job.media:
                    if m.metadata_b64 and json_b64_decode(m.metadata_b64)[key] == value and m.name == video_name:
                        info(f'Found media matching {video_name} and {key} {value} in job {job.name}')
                        media = m
                        break
    if not media:  # can't find by metadata, try by name
        media = [m for m in job.media if m.name == video_name]

    if media:
        if len(media) > 0:
            media = media[0]

        info(f'Found media {video_name} in job {job.name}')

        if status == Status.QUEUED and media.status == Status.RUNNING or media.status == Status.SUCCESS or media.status == Status.FAILED:
            info(f'Media {video_name} in job {job.name} is already {media.status}. Not updating to {status}')
            return

        # Update the media status, timestamp and any additional kwargs
        media.status = status
        media.updatedAt = datetime.utcnow()

        # add metadata if there was one in the kwargs
        if 'metadata_b64' in kwargs:
            media.metadata_b64 = kwargs['metadata_b64']
        else:
            media.metadata_b64 = json_b64_encode(kwargs)

        # Update the metadata
        metadata_json = json_b64_decode(media.metadata_b64)
        for key, value in kwargs.items():
            if key in metadata_json:
                metadata_json[key] = value
        media.metadata_b64 = json_b64_encode(metadata_json)

        db.merge(media)

    else:
        info(f'A new media {video_name} was added to job {job.name} kwargs {kwargs}')
        new_media = Media(name=video_name,
                          status=status,
                          job=job,
                          metadata_b64=json_b64_encode(kwargs),
                          updatedAt=datetime.utcnow())
        db.add(new_media)
        job.media.append(new_media)



# deepsea-ai, Apache-2.0 license
# Filename: database/job/database.py
# Description: Job database

from typing import List

from pydantic_sqlalchemy import sqlalchemy_to_pydantic
from sqlalchemy import Column, ForeignKey, Integer, String, create_engine, func, TIMESTAMP
from sqlalchemy.orm import relationship, sessionmaker, declarative_base

from deepsea_ai.config.config import Config
from deepsea_ai.database.job.misc import JobType, Status
from deepsea_ai.logger import info

Base = declarative_base()


class JobBase(Base):
    __abstract__ = True

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    engine = Column(String, nullable=False)
    job_type = Column(String, nullable=False, default=JobType.SAGEMAKER)
    createdAt = Column(TIMESTAMP(timezone=True),
                       nullable=False, server_default=func.now())


class MediaBase(Base):
    __abstract__ = True

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False)
    status = Column(String, nullable=False, default=Status.UNKNOWN)
    metadata_b64 = Column(String, nullable=True)
    createdAt = Column(TIMESTAMP(timezone=True),
                       nullable=False, server_default=func.now())
    updatedAt = Column(TIMESTAMP(timezone=True),
                       default=None, onupdate=func.now())

class Job(JobBase):
    __tablename__ = "job"

    # one-to-many relationship with the Media table
    media = relationship("Media", backref="job", passive_deletes=True)


class Media(MediaBase):
    __tablename__ = "media"

    job_id = Column(Integer, ForeignKey("job.id", ondelete="CASCADE"))


PydanticJob = sqlalchemy_to_pydantic(Job)
PydanticMedia = sqlalchemy_to_pydantic(Media)


class PydanticJobWithMedias(PydanticJob):
    media: List[PydanticMedia] = []


def init_db(cfg: Config, reset: bool = False) -> sessionmaker:
    """
    Initialize the job cache database
    :param cfg: The configuration
    :param reset: Whether to reset the database
    :return: A sessionmaker
    """
    job_db_path = cfg.job_db_path
    account = cfg('aws', 'account_id')

    # Create the output path to store the database if it doesn't exist
    job_db_path.mkdir(parents=True, exist_ok=True)

    # Name the database based on the account number to avoid collisions
    db = f'sqlite_job_cache_{account}.db'
    info(f"Initializing job cache database in {job_db_path} as {db}")
    engine = create_engine(f"sqlite:///{db}", connect_args={"check_same_thread": True}, echo=False)

    Base.metadata.create_all(engine, tables=[Job.__table__, Media.__table__])

    if reset:
        # Clear the database
        with sessionmaker(bind=engine).begin() as db:
            db.query(Job).delete()
            db.query(Media).delete()

    return sessionmaker(bind=engine)



# deepsea-ai, Apache-2.0 license
# Filename: database/job/misc.py
# Description: Misc. job database functions

import hashlib


class Status:
    QUEUED = "QUEUED"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"


class JobType:
    SAGEMAKER = "SAGEMAKER"
    ECS = "ECS"
    DOCKER = "DOCKER"


def job_hash(job: str) -> str:
    """
    Hash the job name and cluster to create a unique identifier for the job
    """
    md5val = hashlib.md5(job.encode('latin')).hexdigest()
    return f"{md5val[:8]}-{md5val[8:12]}-{md5val[12:16]}-{md5val[16:20]}-{md5val[20:]}".upper()



# deepsea-ai, Apache-2.0 license
# Filename: commands/process.py
# Description: Process a collection of videos with the SageMaker ScriptProcessor

import os
import inspect
import uuid

import boto3
import json
from datetime import datetime
from pathlib import Path
from sqlalchemy.orm import Session, sessionmaker
from deepsea_ai.config import config as cfg
from deepsea_ai.commands.upload_tag import get_prefix
from deepsea_ai.database.job.database import Job, Media, PydanticJobWithMedias
from deepsea_ai.database.job.database_helper import update_media, json_b64_encode
from deepsea_ai.database.job.misc import Status, JobType
from deepsea_ai.logger import debug, info, err

from sagemaker.processing import ScriptProcessor, ProcessingInput, ProcessingOutput

code_path = Path(os.path.abspath(inspect.getfile(inspect.currentframe())))


def script_processor_run(session_maker: sessionmaker, dry_run: bool, input_s3: tuple, output_s3: tuple, model_s3: tuple,
                         volume_size_gb: int, instance_type: str, custom_config: cfg.Config,
                         tags: dict, config_s3: str, args: str):
    """
    Process a collection of videos with the ScriptProcessor
    """
    user_name = "SIM" if dry_run else custom_config.get_username()

    arguments = ['dettrack', f"--model-s3=s3://{model_s3.netloc}/{model_s3.path.lstrip('/')}"]
    if args:
        # args needs to be quoted
        args_quoted = f'"{args}"'
        arguments.append(f'--args={args_quoted}')

    if config_s3:
        arguments.append(f'--config-s3={config_s3}')

    info(f'Running with args {arguments}')

    # Construct the uri from the config, e.g.
    # mbari/deepsea-yolov5:1.1.2 => 872338704006.dkr.ecr.us-west-2.amazonaws.com/deepsea-yolov5:1.1.2
    account = "00000" if dry_run else custom_config.get_account()
    region = "us-west-2" if dry_run else custom_config.get_region()
    image_uri_docker = custom_config('docker', 'strongsort_container')
    image_uri_ecr = f"{account}.dkr.ecr.{region}.amazonaws.com/{image_uri_docker}"
    # log the video as running; the processor is the docker image
    processor = image_uri_ecr.split('/')[-1]
    base_job_name = f'strongsort-yolov5-{user_name}'

    script_processor = ScriptProcessor(command=['python3'],
                                       image_uri=image_uri_ecr,
                                       role=custom_config.get_role(),
                                       instance_count=1,
                                       base_job_name=base_job_name,
                                       instance_type=instance_type,
                                       volume_size_in_gb=volume_size_gb,
                                       max_runtime_in_seconds=172800,
                                       tags=tags)

    # log it
    info(f"Start script processor for inputs s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}")
    job_name = f"{base_job_name}-{datetime.utcnow().strftime('%Y%m%dT%H%M%S')}"
    info(f"Job name: {job_name}")

    def log_fini(db: Session, j: Job, status: str, **kwargs):
        j_p = PydanticJobWithMedias.from_orm(j)
        for m in j_p.media:
            update_media(db, j, m.name, status, **kwargs)

    if not dry_run:
        # get a list of videos in the input bucket
        s3 = boto3.resource('s3')
        bucket = s3.Bucket(input_s3.netloc)
        videos = [obj.key for obj in bucket.objects.filter(Prefix=input_s3.path.lstrip('/'))]
        debug(f'Found videos {videos}')

        # Don't continue if there are no videos
        if len(videos) == 0:
            msg = f"No videos found in s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}"
            err(msg)
            return

        video_names = [Path(v).name for v in videos]

        # Add the job to the database and set the status to QUEUED for each video
        job = Job(engine=processor,
                  name=job_name,
                  job_type=JobType.SAGEMAKER)

        for name in video_names:
            m = Media(name=name,
                      status=Status.QUEUED,
                      updatedAt=datetime.now(),
                      job=job,
                      metadata_b64=json_b64_encode({
                          'image_uri_ecr': image_uri_ecr,
                          'instance_type': instance_type,
                          'volume_size_in_gb': volume_size_gb,
                          'max_runtime_in_seconds': 172800,
                          'tags': tags,
                          'arguments': arguments,
                          'processing_job_arn': '',
                          'error': ''
                      }))
            job.media.append(m)

        with session_maker.begin() as db:
            db.add(job)

        # Run the script processor
        script_processor.run(code=f'{code_path.parent.parent.parent}/deepsea_ai/pipeline/run_strongsort.py',
                             arguments=arguments,
                             inputs=[ProcessingInput(
                                     source=f"s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}",
                                     destination='/opt/ml/processing/input')],
                             outputs=[ProcessingOutput(source='/opt/ml/processing/output',
                                                       destination=f"s3://{output_s3.netloc}/{output_s3.path.lstrip('/')}")]
                             )

        # log success/failure
        with session_maker.begin() as db:
            job = db.query(Job).filter(Job.name == job_name).first()
            if script_processor.jobs[-1].describe()['ProcessingJobStatus'] == 'Failed':
                reason = script_processor.jobs[-1].describe()['FailureReason']
                msg = f"Script processor failed for inputs s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}: {reason}"
                log_fini(db, job, Status.FAILED, processing_job_arn=script_processor.jobs[-1].job_name, error=reason)
                err(msg)
                raise Exception(msg)
            else:
                debug(f"Script processor succeeded for inputs s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}")
                log_fini(db, job, Status.SUCCESS, processing_job_arn=script_processor.jobs[-1].job_name)
    else:
        debug(f"Script processor dry run for inputs s3://{input_s3.netloc}/{input_s3.path.lstrip('/')}")


def batch_run(session_maker: sessionmaker, resources: dict, video_path: Path, job_name: str, user_name: str, clean: bool, args: str):
    """
    Process a collection of videos in with a cluster in the Elastic Container Service [ECS]
    """
    # the queue to submit the processing message to
    queue_name = resources['VIDEO_QUEUE']

    # Get the service resource
    sqs = boto3.resource('sqs')

    prefix_path = get_prefix(video_path)

    # Setup message dict
    # Generate a unique uuid to track the video processing
    message_uuid = str(uuid.uuid4())
    message_dict = {"video": f"{prefix_path}/{video_path.name}",
                    "clean": "True" if clean else "False",
                    "user_name": user_name,
                    "metadata_b64": json_b64_encode({"message_uuid": message_uuid}),
                    "job_name": job_name}

    # If args are provided, add them to the message dict
    if args:
        # Strip off the quotes
        args = args.strip('"')
        message_dict["args"] = args

    queue = sqs.get_queue_by_name(QueueName=queue_name)
    json_object = json.dumps(message_dict, indent=4)

    # Create a message group based on the time and the video name
    group_id = f"{datetime.utcnow().strftime('%Y%m%dT%H%M%S')}-{video_path.name}"

    # Create a new message
    response = queue.send_message(MessageBody=json_object, MessageGroupId=group_id)
    info(f"Message queued to {queue_name}. MessageId: {response.get('MessageId')}")

    with session_maker.begin() as db:
        # Add the job to the database if it doesn't exist
        job = db.query(Job).filter(Job.name == job_name).first()
        try:
            if job is None:
                job = Job(engine=resources['CLUSTER'],
                          name=job_name,
                          job_type=JobType.ECS)
                db.add(job)
        except Exception as ex:
            err(f"Failed to add job {job_name} to cache: {ex}")
            raise ex

    with session_maker.begin() as db:
        job = db.query(Job).filter(Job.name == job_name).first()
        info(f"Added job {job.name} running on {resources['CLUSTER']} to cache.")
        update_media(db, job, f"{prefix_path}/{video_path.name}", Status.QUEUED, message_uuid=message_uuid)


# deepsea-ai, Apache-2.0 license
# Filename: common_args.py
# Description: Common arguments for processing commands

import click

# Common arguments for processing commands
job_option = click.option('--job', type=str, required=True,
                          help='Name of the job, e.g. DiveV4361 benthic outline')
cluster_option = click.option('--cluster', type=str, required=True,
                              help='Name of the cluster to use to batch process. '
                                   'This must correspond to an available Elastic '
                                   'Container Service cluster. Clusters names correspond '
                                   'to available model names')
dry_run_option = click.option('--dry-run', is_flag=True, default=False,
                              help='Run the command without actually submitting the job.')
config_s3_option = click.option('--config-s3', type=str,
                                help='S3 location of tracking algorithm config yaml file')
args = click.option('--args', type=str, help='Arguments to pass directly to the docker image ')



# deepsea-ai, Apache-2.0 license
# Filename: commands/upload_tag.py
# Description: Bucket upload and tagging utility

import botocore
import boto3
import time
from pathlib import Path
from urllib.parse import urlparse
from deepsea_ai.logger import info, err, critical, exception

from . import bucket


def video_data(videos: list[Path], input_s3: tuple, tags: dict, dry_run: bool = False):
    """
     Does an upload and tagging of a collection of videos to S3
    :param videos: Array of video files in the input_path to upload
    :param input_s3: Base bucket to upload to, e.g. 902005-video-in-dev
    :param tags: Tags to assign to the video
    :param dry_run: If true, do not upload or tag
    :return: Uploaded bucket path, Size in GB of video data
    """
    if dry_run:
        for v in videos:
            info(f'dry-run: Would have uploaded {v.as_posix()} to {input_s3.netloc} with tags {tags}')
        return urlparse(f's3://{input_s3.netloc}/{get_prefix(videos[0])}/', allow_fragments=True), 0

    s3 = boto3.client('s3')
    s3_resource = boto3.resource('s3')

    # upload and tag the video objects individually
    uploaded_videos = []
    size_gb = 1
    for v in videos:
        # add the size of the video to the total
        size_gb += v.stat().st_size /(1024**3)
        prefix_path = get_prefix(v)

        if input_s3.path:
            target_prefix = f"{input_s3.path}/{prefix_path.lstrip('/')}/{v.name}"
        else:
            target_prefix = f"{prefix_path.lstrip('/')}/{v.name}"

        info(f'Checking {v} in s3://{input_s3.netloc}/{target_prefix}...')
        uploaded_videos.append(f"s3://{input_s3.netloc}/{target_prefix}")

        # check if the video exists in s3
        try:
            s3_resource.Object(input_s3.netloc, target_prefix).load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "403":
                info(f'Found s3://{input_s3.netloc}/{target_prefix} but do not have permission to access it. Continuing...')
                continue
            if e.response['Error']['Code'] == "404":
                upload_success = False
                # The video does not exist so upload and retry
                for retry in range(10):
                    try:
                        info(f'Uploading {v} to s3://{input_s3.netloc}/{target_prefix}...')
                        s3.upload_file(v.as_posix(), input_s3.netloc, target_prefix)
                        upload_success = True
                        info(f'File {v} uploaded successfully')
                        break
                    except FileNotFoundError:
                        info(f"Local file '{v}' not found.")
                        exception(f"Error uploading {v} to s3.")
                        time.sleep(60)
                    except botocore.exceptions.EndpointConnectionError as e:
                        info(f'Network error: {e} Retrying every 60 seconds...')
                        time.sleep(60)
                    except TimeoutError as e:
                        exception(f'Timeout error {e}. Retrying every 60 seconds...')
                        time.sleep(60)
                    except Exception as e:
                        exception(f'Error uploading {e} {v} to s3. Retrying every 60 seconds...')
                        time.sleep(60)

                if not upload_success:
                    critical(f"Error uploading {v} to s3 after {retry} retries. Aborting.")
                    raise Exception(f"Error uploading {v} to s3 after {retry} retries. Aborting.")
            else:
                exception(e)
                raise
        else:
            # the video does exist.
            info(f'Found s3://{input_s3.netloc}/{target_prefix} ...skipping upload')

        try:
            # tag it
            info(f'Tagging {v} with {tags}...')
            s3.put_object_tagging(Bucket=input_s3.netloc, Key=f'{target_prefix}', Tagging={'TagSet': tags})
        except Exception as error:
            raise error

    # If only one video was uploaded, return the video path
    if len(uploaded_videos) == 1:
        return urlparse(uploaded_videos[0], allow_fragments=True), size_gb
    else:
        # Otherwise, return the prefix path
        return urlparse(f"s3://{input_s3.netloc}/{get_prefix(videos[0])}/", allow_fragments=True), size_gb


def training_data(data: [Path], input: tuple, tags: dict, training_prefix: str):
    """
     Does an upload and tagging of training data to S3
    :param data: Paths to training data to upload
    :param input: Bucket to upload to
    :param bucket: Tags to assign to the video
    :param training_prefix: Training prefix to append to the bucket upload
    :return: Uploaded bucket path, Size in GB of training data
    """

    # upload and tag the video objects individually
    s3 = boto3.client('s3')
    s3_resource = boto3.resource('s3')

    for d in data:
        if not d.exists():
            err(f"Error: {d} does not exist")
            exit(-1)

    for d in data:
        # check if the data exists in s3
        # all the data needs to be under the same prefix for training
        target_prefix = f'{training_prefix}/{d.name}'
        try:
            s3_resource.Object(input.netloc, target_prefix).load()
        except botocore.exceptions.ClientError as e:
            info(f'{e} {d} does not exist in s3://{input.netloc}/{target_prefix}. Uploading...')
            if e.response['Error']['Code'] == "404":
                # The data does not exist so upload
                try:
                    with open(d.as_posix(), "rb") as f:
                        info(f'Uploading {d} to s3://{input.netloc}/{target_prefix}...')
                        s3.upload_fileobj(f, input.netloc, target_prefix)
                except Exception as error:
                    error(f"Error {error} uploading to s3")
            else:
                raise
        else:
            # the data already exist so skip over it
            info(f'Found s3://{input.path}/{target_prefix} ...skipping upload')

        try:
            # tag it
            s3.put_object_tagging(Bucket=input.netloc, Key=target_prefix, Tagging={'TagSet': tags})
        except Exception as error:
            exception(error)
            raise error

    output = urlparse(f"s3://{input.netloc}/{training_prefix}/")
    size_gb = bucket.size(output)

    return output, size_gb


def get_prefix(path: Path):
    """
    Get the prefix from a path, stripping away any volume or drive information
    :param path: Path to get the prefix from
    :return: Prefix
    """
    prefix = None
    for m in ['Volumes/', 'mnt/', 'Users/', 'home/']:
        if m in path.as_posix():
            prefix = path.parent.as_posix().split(m)[-1].lstrip('/')
            break

    # Do not allow empty prefixes
    if not prefix:
        prefix = '/'
    return prefix



# deepsea-ai, Apache-2.0 license
# Filename: commands/monitor_utils.py
# Description: Helper commands to monitor the status of tasks run on an ECS cluster.

import json
from datetime import datetime
from pathlib import Path

import boto3
from botocore.exceptions import ClientError
from sqlalchemy.orm import Session, sessionmaker

from deepsea_ai.database.job import Job, Status
from deepsea_ai.database.job.database_helper import update_media, json_b64_decode
from deepsea_ai.logger import info, err, exception, debug


def log_scaling_activities(resources: dict, num_records: int = 10) -> int:
    """
    Log scaling activities for a given resource
    :param resources: Dictionary of resources
    :param num_records: Number of records to return
    :param report: If true, log the activities
    :return: Number of activities
    """
    client = boto3.client('autoscaling')
    response = client.describe_scaling_activities(
        ActivityIds=[],
        AutoScalingGroupName=resources['ASG'],
        IncludeDeletedGroups=False,
        MaxRecords=num_records)

    for i in response['Activities']:
        msg = f'{i["StartTime"]} {i["Description"]}  {i["Cause"]}'
        info(msg)

    return len(response['Activities'])


def parse_message(message: dict) -> dict:
    """
    Parse the JSON formated message from the queue body
    :param message: Message from the queue
    :return: Dictionary of the message with the timestamp added
    """
    try:
        if 'Body' not in message:
            err(f'No body in message {message}')
            return None
        b = json.loads(message['Body'])
        utc_secs = int(message['Attributes']['ApproximateFirstReceiveTimestamp'])
        timestamp = datetime.utcfromtimestamp(utc_secs / 1000)
        b['timestamp'] = timestamp.strftime('%Y%m%dT%H%M%S')
        return b
    except Exception as e:
        exception(e)
        return None


def fetch_and_parse(client, q):
    """
    Fetch queues with long polling
    Can only fetch 10 messages at a time; wait up to 20 seconds for a message
    :param client: the sqs client
    :param q: The queue to fetch messages from
    :return: List of parsed messages from the queue
    """
    messages = []
    while True:
        response = client.receive_message(QueueUrl=q,
                                          MaxNumberOfMessages=10,
                                          WaitTimeSeconds=20,
                                          MessageAttributeNames=['All'],
                                          AttributeNames=['All'])
        if 'Messages' not in response:
            break

        for m in response['Messages']:
            message = parse_message(m)
            if message:
                messages.append(message)

    return messages


def update_job(db: Session, sqs_message: dict, cluster: str, status: str):
    """
    Helper function to update the database
    :param db: Database session
    :param sqs_message: The message from the queue
    :param cluster: The cluster the job is running on
    :param status: The status of the video, either QUEUED, SUCCESS, or FAILED
    """
    job = db.query(Job).filter_by(name=sqs_message['job_name'], engine=cluster).first()

    if job is None:
        err(f'Job {sqs_message["job_name"]} not found in database.')
        # Add the job to the database
        job = Job(name=sqs_message['job_name'], engine=cluster, job_type='ECS')
        db.add(job)
    else:
        info(f'Found job {job.name} running on {cluster} in cache.')

    # pass through the metadata
    update_media(db, job, sqs_message["video"], status, metadata_b64=sqs_message['metadata_b64'])


def log_queue_status(session_maker: sessionmaker, resources: dict) -> dict:
    """
    Logs the status of the queues
    :param resources: Dictionary of resources
    :return: Dictionary of the number of messages in each queue
    :param resources: Dictionary of resources in the cluster
    :return: Dictionary of the number of messages visible in each queue
    """
    client = boto3.client('sqs')
    # queues = ['VIDEO_QUEUE', 'TRACK_QUEUE', 'DEAD_QUEUE']
    queues = ['DEAD_QUEUE']
    cluster = resources['CLUSTER']
    processor = resources['PROCESSOR']
    num_messages_visible = {}
    num_messages_invisible = {}

    try:
        for q in queues:
            response = client.get_queue_attributes(
                QueueUrl=resources[q],
                AttributeNames=['ApproximateNumberOfMessages', 'ApproximateNumberOfMessagesNotVisible'])
            num_messages_visible[q] = response['Attributes']['ApproximateNumberOfMessages']
            num_messages_invisible[q] = response['Attributes']['ApproximateNumberOfMessagesNotVisible']

            if q == 'TRACK_QUEUE':
                info(f'{processor}:{q} number of processed videos: '
                     f'{response["Attributes"]["ApproximateNumberOfMessages"]}')
                messages = fetch_and_parse(client, resources[q])
                if messages:
                    with session_maker() as db:
                        for message in messages:
                            update_job(db, message, cluster, Status.SUCCESS)

            if q == 'VIDEO_QUEUE':
                info(f'{processor}:{q} number of videos to process: '
                     f'{response["Attributes"]["ApproximateNumberOfMessages"]} ')
                info(f' number of videos in progress: '
                     f'{response["Attributes"]["ApproximateNumberOfMessagesNotVisible"]}')

                # This needs to be tested in a multiple user scenario
                # For now, assume single user, single command execution use-case only
                # messages = fetch_and_parse(client, resources[q])
                # if messages:
                #     with session_maker() as db:
                #         for message in messages:
                #             update_job(db, message, cluster, Status.QUEUED)

            if q == 'DEAD_QUEUE':
                info(f'{processor}:{q} number of failed videos: '
                     f'{response["Attributes"]["ApproximateNumberOfMessages"]}')
                messages = fetch_and_parse(client, resources[q])
                if messages:
                    with session_maker() as db:
                        for message in messages:
                            update_job(db, message, cluster, Status.FAILED)

    except ClientError as e:
        exception(e)
        raise e

    return num_messages_visible


def receive_messages(queue, max_number, wait_time):
    """
    Receive a batch of messages in a single request from an SQS queue.

    :param queue: The queue from which to receive messages.
    :param max_number: The maximum number of messages to receive. The actual number
                       of messages received might be less.
    :param wait_time: The maximum time to wait (in seconds) before returning. When
                      this number is greater than zero, long polling is used. This
                      can result in reduced costs and fewer false empty responses.
    :return: The list of Message objects received. These each contain the body
             of the message and metadata and custom attributes.
    """
    try:
        messages = queue.receive_messages(
            MessageAttributeNames=['All'],
            MaxNumberOfMessages=max_number,
            WaitTimeSeconds=wait_time
        )
        for msg in messages:
            debug(f"Received message: {msg.message_id}: {msg.body}")
    except ClientError as error:
        exception(f"Couldn't receive messages from queue: {queue}")
        raise error
    else:
        return messages
